--[[	SlashHelp
	by SDPhantom
	http://www.phantomweb.org	]]
------------------------------------------

local name,addon=...;

--------------------------
--[[	Local Pointers	]]
--------------------------
local _G=_G;
local math=math;
local table=table;
local ipairs=ipairs;
local pairs=pairs;
local rawget=rawget;
local rawset=rawset;
local tonumber=tonumber;
local type=type;
local IsSecureCmd=IsSecureCmd;
local ChatTypeInfo=ChatTypeInfo;
local SlashCmdList=SlashCmdList;

--------------------------
--[[	Metatables	]]
--------------------------
local CIKeyMeta={--	Forces keys to be case-insensitive
	__index=function(t,k) return rawget(t,type(k)=="string" and k:lower() or k); end;
	__newindex=function(t,k,v) return rawset(t,type(k)=="string" and k:lower() or k,v); end;
}

--------------------------
--[[	Cache Vars	]]
--------------------------
--	Filter cache
local FilterCache={};
local LastFilter="";

--	Scanner cache
local AddOnInfo={};
local SlashInfo={};
local SlashGroupLookup={};
local SlashOrder={};
local GroupOrder={};

setmetatable(AddOnInfo,CIKeyMeta);
setmetatable(SlashInfo,CIKeyMeta);

--	Progressive scanner queue
local ScanCache={};

--------------------------
--[[	Local Funcs	]]
--------------------------
local function RegisterSlashCommand(slash,token,cmdtype,addonid)
	local addon=AddOnInfo[addonid or ""];
	if not addon and (addonid or "")~="" then
		local id,title,notes=GetAddOnInfo(addonid);
		local author=GetAddOnMetadata(addonid,"Author");
		local version=GetAddOnMetadata(addonid,"Version");
		if id then
			addon={
				ID=id;
				Name=title;
				Version=version;
				Author=author;
				Notes=notes;
			};
			AddOnInfo[addonid]=addon;
		end
	end
	
	local aliases=SlashGroupLookup[token];
	if not aliases then
		aliases={};
		SlashGroupLookup[token]=aliases;
		GroupOrder[#GroupOrder+1]=token;
	end

	if type(slash)=="table" then
		for i,j in ipairs(slash) do
			if not SlashInfo[j] then
				aliases[#aliases+1]=j;
				SlashOrder[#SlashOrder+1]=j;
				SlashInfo[j]={
					Slash=j;
					Token=token;
					Type=cmdtype;
					Aliases=aliases;
					AddOn=addon;
				}
			end
		end
	elseif not SlashInfo[slash] then
		aliases[#aliases+1]=slash;
		SlashOrder[#SlashOrder+1]=slash;
		SlashInfo[slash]={
			Slash=slash;
			Token=token;
			Type=cmdtype;
			Aliases=aliases;
			AddOn=addon;
		}
	end
end

local function SlashSortFunc(v1,v2)
	s1,s2=(type(v1)=="string" and v1 or tostring(v1 or "")):lower(),(type(v2)=="string" and v2 or tostring(v2 or "")):lower();
	local l1,l2=#v1,#v2;
	local len=math.min(l1,l2);
	if len<=0 then
		return l1<l2;
	else
		for i=1,len do
--			Following code checks letter before case (A < a < B < b < C < c < etc)
			local b1,b2=s1:byte(i),s2:byte(i);
			if b1~=b2 then return b1<b2;--	Check lower() conversion
			else
				b1,b2=v1:byte(i),v2:byte(i);--	Check case (upper before lower)
				if b1~=b2 then return b1<b2; end
			end
		end
		return l1<l2;
	end
end

local function SortSlashCommands()
	table.sort(SlashOrder,SlashSortFunc);
	table.sort(GroupOrder,SlashSortFunc);
	for i,j in pairs(SlashGroupLookup) do table.sort(j,SlashSortFunc); end
end

--------------------------
--[[	Filter Func	]]
--------------------------
function addon:BuildFilter(filter)
	filter=(filter or LastFilter):lower();
	table.wipe(FilterCache);

	if filter:find("^/") then
		if self.Options.SortType=="Group" then--	Group sort
			for i,j in ipairs(GroupOrder) do
				for k,l in ipairs(SlashGroupLookup[j]) do
					local ptr=SlashInfo[l];
					if filter==l:sub(1,math.min(#filter,#l)):lower() and self.Options.Show[ptr.Type] then
						FilterCache[#FilterCache+1]=ptr;
					end
				end
			end
		else--						Slash sort (default)
			for i,j in ipairs(SlashOrder) do
				local ptr=SlashInfo[j];
				if filter==j:sub(1,math.min(#filter,#j)):lower() and self.Options.Show[ptr.Type] then
					FilterCache[#FilterCache+1]=ptr;
				end
			end
		end
	end

	LastFilter=filter;
	return FilterCache;
end

--------------------------
--[[	Scanner Funcs	]]
--------------------------
local function InitialScan()
	for i,j in pairs(_G) do
		if type(j)=="string" and j:find("^/") then
			if not SlashInfo[j] then
				local token,isemote=i:match("^SLASH_(.-)%d+$"),false;
				if not token then token,isemote=_G["EMOTE"..(i:match("^EMOTE(%d+)_CMD%d+$") or 0).."_TOKEN"],true; end

				if token then
					local cmdtype;

					if isemote then			cmdtype="Emote";
					elseif IsSecureCmd(j) then	cmdtype="Secure";
					elseif ChatTypeInfo[token] then	cmdtype="Chat"
					elseif SlashCmdList[token] then	cmdtype="Slash";
					end

					if cmdtype then
						local secure,id;
						if cmdtype=="Slash" then secure,id=issecurevariable(i); end
						RegisterSlashCommand(j,token,cmdtype,((not secure and id~="") and id or nil));
					end
				end
			end
		end
	end

	SortSlashCommands();
	addon:BuildFilter(nil);
end

local function ProgressiveScan()
	if #ScanCache>0 then
		for i,j in ipairs(ScanCache) do
			local cmdlist={};
			local k,str=1,_G["SLASH_"..j..1];
			while str do
				local secure,id=issecurevariable("SLASH_"..j..k);
				RegisterSlashCommand(str,j,"Slash",((not secure and id~="") and id or nil));
				k=k+1;str=_G["SLASH_"..j..k];--	k needs to be incremented before referenced for str
			end

		end
		table.wipe(ScanCache);

		SortSlashCommands();
		addon:BuildFilter(nil);
	end
end

--------------------------------------------------
--[[	Progressive Scanner Trigger Hook	]]
--------------------------------------------------
do	local oldmeta,newindexfunc=getmetatable(SlashCmdList),function(t,k,v) rawset(t,k,v);ScanCache[#ScanCache+1]=k; end;
	if oldmeta then
		if oldmeta.__newindex then
			hooksecurefunc(oldmeta,"__newindex",newindexfunc);
		else
			oldmeta.__newindex=newindexfunc;
		end
	else
		setmetatable(SlashCmdList,{__newindex=newindexfunc});
	end
	addon.EventFrame:HookScript("OnUpdate",function() if #ScanCache>0 then ProgressiveScan(); end end);
end

----------------------------------
--[[	Initial Scanner Event	]]
----------------------------------
addon.EventFrame:RegisterEvent("ADDON_LOADED");
addon.EventFrame:HookScript("OnEvent",function(self,event,...) if event=="ADDON_LOADED" and (...)==name then InitialScan() end end);
